#include <cryptoauthlib.h>
#include "basic/atca_basic.h"

#define BOOL_TO_ONOFF(b) (b ? "on" : "off")
#define NIBBLE_TO_HEX_CHAR(i) ((i <= 9) ? ('0' + i) : ('A' - 10 + i))
#define HIGH_NIBBLE(i) ((i >> 4) & 0x0F)
#define LOW_NIBBLE(i) (i & 0x0F)

#define HEX_CHAR_TO_NIBBLE(c) ((c >= 'A') ? (c - 'A' + 0x0A) : (c - '0'))
#define HEX_PAIR_TO_BYTE(h, l) ((HEX_CHAR_TO_NIBBLE(h) << 4) + HEX_CHAR_TO_NIBBLE(l))

#define IS_HEX_CHAR(c) ( (((c >= 'A') && (c <= 'F')) || ((c >= '0') && (c <= '9'))) ? true : false)
#define ARRAY_SIZE(a) (sizeof(a)/sizeof(a[0]))

#define debugSerial SerialUSB
#define debugBaud 57600
#define debugDelay 10000

void setup() 
{
  while ((!debugSerial) && (millis() < debugDelay)){
    // Wait debugDelay milliseconds for debugSerial to open
  }

  // Start debug stream
  debugSerial.begin(debugBaud);
  debugSerial.println("Starting...");
  debugSerial.println();
  
  //Init chip
  cfg_ateccx08a_i2c_default.iface.atcai2c.bus = 0;
  debugSerial.print("Initialising ATECCX08A:...");
  showResult(atcab_init(&cfg_ateccx08a_i2c_default));
  debugSerial.println();

  //Read serial number
  uint8_t serial_num[ATCA_SERIAL_NUM_SIZE];
  debugSerial.print("Reading serial number:...");
  showResult(atcab_read_serial_number(serial_num));
  printHex(serial_num, sizeof(serial_num), sizeof(serial_num));
  debugSerial.println();

  //Read revision information
  uint8_t rev_info[4];
  debugSerial.print("Reading revision info:...");
  showResult(atcab_info(rev_info));
  printHex(rev_info, sizeof(rev_info), sizeof(rev_info));
  debugSerial.println();

  //Read configuration lock status
  bool isLocked = false;
  debugSerial.print("Reading configuration lock status:...");
  showResult(atcab_is_locked(LOCK_ZONE_CONFIG, &isLocked));
  debugSerial.print("The configuration zone is:...");
  debugSerial.println((isLocked ? "LOCKED" : "UNLOCKED"));
  debugSerial.println();

  if (isLocked) {
    //TRNG test
    uint8_t random_num[32];
    debugSerial.print("Requesting TRNG:...");
    showResult(atcab_random(random_num));
    printHex(random_num, sizeof(random_num), 16);
    debugSerial.println();
  }
  else {
    debugSerial.println("Device is must first be configured...");
    debugSerial.println();
  }
}

void loop() 
{ 
}

void printHex(const uint8_t* buff, uint8_t len, uint8_t width)
{
    char strA[16];
    for (uint8_t i = 0; i < len; i++)
    {
        if (i % width == 0)
        {
          sprintf(strA,"0x%04x",i);
          debugSerial.print(String((i == 0 ? "" : "\r\n")) + String(strA) + ": ");
        }
        sprintf(strA,"0x%02X, ",buff[i]);
        debugSerial.print(strA);
        
    }
    debugSerial.println();
}

void printRawHex(const uint8_t* buff, uint8_t len)
{
    for (uint8_t i = 0; i < len; i++)
    {
      debugSerial.print((char)NIBBLE_TO_HEX_CHAR(HIGH_NIBBLE(buff[i])));
      debugSerial.print((char)NIBBLE_TO_HEX_CHAR(LOW_NIBBLE(buff[i])));
    }
    debugSerial.println();
}

void halt()
{
  while (true) {
    delay(1000);
  }
}

void showResult(ATCA_STATUS result)
{
  String message;
  
  switch(result) {
    case ATCA_SUCCESS                 :message = "SUCCESS"; break;
    case ATCA_CONFIG_ZONE_LOCKED      :message = "CONFIG_ZONE_LOCKED"; break;
    case ATCA_DATA_ZONE_LOCKED        :message = "DATA_ZONE_LOCKED"; break;
    case ATCA_WAKE_FAILED             :message = "WAKE_FAILED"; break;
    case ATCA_CHECKMAC_VERIFY_FAILED  :message = "CHECKMAC_VERIFY_FAILED"; break;
    case ATCA_PARSE_ERROR             :message = "PARSE_ERROR"; break;
    case ATCA_STATUS_CRC              :message = "STATUS_CRC"; break;
    case ATCA_STATUS_UNKNOWN          :message = "STATUS_UNKNOWN"; break;
    case ATCA_STATUS_ECC              :message = "STATUS_ECC"; break;
    case ATCA_FUNC_FAIL               :message = "FUNC_FAIL"; break;
    case ATCA_GEN_FAIL                :message = "GEN_FAIL"; break;
    case ATCA_BAD_PARAM               :message = "BAD_PARAM"; break;
    case ATCA_INVALID_ID              :message = "INVALID_ID"; break;
    case ATCA_INVALID_SIZE            :message = "INVALID_SIZE"; break;
    case ATCA_RX_FAIL                 :message = "RX_FAIL"; break;
    case ATCA_RX_NO_RESPONSE          :message = "RX_NO_RESPONSE"; break;
    case ATCA_RESYNC_WITH_WAKEUP      :message = "RESYNC_WITH_WAKEUP"; break;
    case ATCA_PARITY_ERROR            :message = "PARITY_ERROR"; break;
    case ATCA_TX_TIMEOUT              :message = "TX_TIMEOUT"; break;
    case ATCA_RX_TIMEOUT              :message = "RX_TIMEOUT"; break;
    case ATCA_COMM_FAIL               :message = "COMM_FAIL"; break;
    case ATCA_TIMEOUT                 :message = "TIMEOUT"; break;
    case ATCA_BAD_OPCODE              :message = "BAD_OPCODE"; break;
    case ATCA_WAKE_SUCCESS            :message = "WAKE_SUCCESS"; break;
    case ATCA_EXECUTION_ERROR         :message = "EXECUTION_ERROR"; break;
    case ATCA_UNIMPLEMENTED           :message = "UNIMPLEMENTED"; break;
    case ATCA_ASSERT_FAILURE          :message = "ASSERT_FAILURE"; break;
    case ATCA_TX_FAIL                 :message = "TX_FAIL"; break;
    case ATCA_NOT_LOCKED              :message = "NOT_LOCKED"; break;
    case ATCA_NO_DEVICES              :message = "NO_DEVICES"; break;
  }

  debugSerial.println(message);

  if (result != ATCA_SUCCESS) {
    halt();
  }
}
